class ChunkExecutorManager(object):
    def __init__(self):
        self.chunk_executors = {}

    def register(self, executor_id: str, executor_func: callable):
        self.chunk_executors.update({ executor_id: executor_func })
        return self

    def get(self, executor_id: str):
        if executor_id in self.chunk_executors:
            return self.chunk_executors[executor_id]
        else:
            return None


import functools

def create_connection_symbol(chunk, handle_name):
  """生成 chunk 连接字符"""
  return f"{get_display_name(chunk)}({handle_name})"

def create_symbol(chunk):
  """生成 chunk 符号"""
  return f"{chunk['id']}({get_display_name(chunk)})"

def get_display_name(chunk):
  return fix_display_text(chunk["title"] or f'chunk-{chunk["id"]}' or 'Unknow chunk')

def fix_display_text(label: str):
  """修复文本展示，避免特殊字符"""
  if not label:
      return label
  escaped_string = label.replace('"', '&quot;')
  return f'"{escaped_string}"'


def deep_copy_simply(obj):
  """简单深拷贝，只处理 dict / list"""
  if isinstance(obj, list):
    return [deep_copy_simply(item) for item in obj]
  elif isinstance(obj, dict):
    return {key: deep_copy_simply(value) for key, value in obj.items()}
  else:
    return obj



import uuid
import yaml
from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace

class Session(ComponentABC):
    def __init__(self, agent):
        self.agent = agent
        self.settings = RuntimeCtxNamespace("plugin_settings.agent_component.Session", self.agent.settings)
        self.current_session_id = None
        self.full_chat_history = []
        self.recent_chat_history = []
        self.abstract = RuntimeCtxNamespace("abstract", self.agent.agent_runtime_ctx)

    def toggle_auto_save(self, is_enabled: bool):
        self.settings.set("auto_save", is_enabled)
        return self.agent

    def toggle_strict_orders(self, is_strict_orders: bool):
        self.settings.set("strict_orders", is_strict_orders)
        return self.agent

    def toggle_manual_chat_history(self, is_manual_chat_history: bool):
        self.settings.set("manual_chat_history", is_manual_chat_history)
        return self.agent

    def set_max_length(self, max_length: int):
        self.settings.set("max_length", max_length)
        return self.agent

    
    def activate(self, session_id: str=None):
        if self.current_session_id != None:
            self.stop()
        if session_id == None or session_id == "":
            self.current_session_id = str(uuid.uuid1())
        else:
            self.current_session_id = session_id
            full_chat_history = self.agent.agent_storage.table("chat_history").get(session_id)
            full_chat_history = full_chat_history if full_chat_history else []
            self.full_chat_history.extend(full_chat_history)
            self.recent_chat_history.extend(full_chat_history)
        return self.current_session_id

    def save(self):
        self.agent.agent_storage.table("chat_history").set(self.current_session_id, self.full_chat_history).save()
        return self.agent

    def stop(self):
        if self.settings.get_trace_back("auto_save", True):
            self.save()
        self.current_session_id = None
        self.full_chat_history = []
        self.recent_chat_history = []
        return self.agent

    def __get_shorten_chat_history(self, current_chat_history: list, max_length: int):
        if len(str(current_chat_history)) > max_length:
            self.__get_shorten_chat_history(current_chat_history[1:], max_length)
        else:
            return current_chat_history

    def shorten_chat_history(self):
        self.recent_chat_history = self.__get_shorten_chat_history(self.recent_chat_history, self.settings.get_trace_back("max_length"))
        return self.agent

    def add_chat_history(self, role: str, content: str):
        if role not in ("user", "assistant"):
            raise Exception("[Agent Component]: Session - add_chat_history() only accept role type 'user' or 'assistant'.")
        is_strict_orders = self.settings.get_trace_back("strict_orders")
        if is_strict_orders:
            if len(self.full_chat_history) > 0:
                if self.full_chat_history[-1]["role"] == role:
                    self.full_chat_history[-1]["content"] += "\n" + content
                    self.recent_chat_history[-1]["content"] += "\n" + content
                else:
                    self.full_chat_history.append({ "role": role, "content": content})
                    self.recent_chat_history.append({ "role": role, "content": content})
            else:
                if role == "user":
                    self.full_chat_history.append({ "role": role, "content": content})
                    self.recent_chat_history.append({ "role": role, "content": content})
                else:
                    self.full_chat_history.extend([
                        { "role": "user", "content": "[EMPTY]" },
                        { "role": role, "content": content},
                    ])
                    self.recent_chat_history.extend([
                        { "role": "user", "content": "[EMPTY]" },
                        { "role": role, "content": content},
                    ])
        else:
            self.full_chat_history.append({ "role": role, "content": content})
            self.recent_chat_history.append({ "role": role, "content": content})
        return self.agent

    def get_chat_history(self, *, is_shorten: bool=False):
        if is_shorten:
            self.recent_chat_history = self.__get_shorten_chat_history(self.recent_chat_history, self.settings.get_trace_back("max_length", 12000))
            return self.recent_chat_history
        else:
            return self.full_chat_history

    def rewrite_chat_history(self, new_chat_history: list):
        self.full_chat_history = new_chat_history.copy()
        self.recent_chat_history = new_chat_history.copy()
        return self.agent

    def __find_input(self, input_data: any):
        if isinstance(input_data, str):
            return input_data
        elif isinstance(input_data, dict):
            if "input" in input_data:
                return str(input_data["input"])
            elif "question" in input_data:
                return str(input_data["question"])
            elif "target" in input_data:
                return str(input_data["target"])
            elif "goal" in input_data:
                return str(input_data["goal"])
            else:
                return yaml.dump(input_data, allow_unicode=True, sort_keys=False)
        else:
            return str(input_data)

    def __find_reply(self, reply_data: any):
        if isinstance(reply_data, str):
            return reply_data
        elif isinstance(reply_data, dict):
            if "reply" in reply_data:
                return str(reply_data["reply"])
            elif "result" in reply_data:
                return str(reply_data["result"])
            elif "response" in reply_data:
                return str(reply_data["response"])
            elif "anwser" in reply_data:
                return str(reply_data["anwser"])
            elif "output" in reply_data:
                return str(reply_data["output"])
            else:
                return yaml.dump(reply_data, allow_unicode=True, sort_keys=False)
        else:
            return str(reply_data)

    def _prefix(self):
        prefix_result = {}
        if self.current_session_id != None and self.current_session_id != "":
            self.shorten_chat_history()
            prefix_result.update({ "chat_history": self.recent_chat_history })
        abstract = self.abstract.get()
        if abstract:
            prefix_result.update({ "abstract": abstract })
        return prefix_result

    def _suffix(self, event: str, data: any):
        is_manual_chat_history = self.settings.get_trace_back("manual_chat_history", False)
        if self.current_session_id != None and self.current_session_id != "" and not is_manual_chat_history:
            if event == "response:finally":
                new_chat_history = [
                    { "role": "user", "content": self.__find_input(data["prompt"]["input"]) },
                    { "role": "assistant", "content": self.__find_reply(data["reply"]) }
                ]
                self.full_chat_history.extend(new_chat_history)
                self.recent_chat_history.extend(new_chat_history)

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": self._suffix,
            "alias": {
                "active_session": { "func": self.activate, "return_value": True },
                "activate_session": { "func": self.activate, "return_value": True },
                "save_session": { "func": self.save },
                "stop_session": { "func": self.stop },
                "toggle_session_auto_save": { "func": self.toggle_auto_save },
                "toggle_strict_orders": { "func": self.toggle_strict_orders },
                "toggle_manual_chat_history": { "func": self.toggle_manual_chat_history },
                "set_chat_history_max_length": { "func": self.set_max_length },
                "add_chat_history": { "func": self.add_chat_history },
                "get_chat_history": { "func": self.get_chat_history, "return_value": True },
                "rewrite_chat_history": { "func": self.rewrite_chat_history },
                "set_abstract": { "func": self.abstract.assign },
            },
        }

def export():
    return ("Session", Session)


import copy

class StorageEditor(object):
    def __init__(
            self,
            *,
            target_data: dict=None,
            no_copy: bool=False,
            get_target_data: callable,
            storage: object,
            table_name: str
        ):
        self.target_data = target_data
        self.no_copy = no_copy
        self.get_target_data = get_target_data
        self.storage = storage
        self.table_name = table_name

    def __locate_pointer(self, keys_with_dots: str):
        if self.target_data == None:
            self.target_data = self.get_target_data() or {}
        keys = keys_with_dots.split('.')
        pointer = self.target_data
        current_key = None
        for key in keys:
            if current_key:
                pointer = pointer[current_key]
            current_key = key
            if key not in pointer:
                pointer[key] = {}
        return pointer, current_key

    def set(self, keys_with_dots: str, value: any):
        pointer, key = self.__locate_pointer(keys_with_dots)
        pointer[key] = value
        return self

    def delta(self, keys_with_dots: str, value: any):
        if isinstance(value, dict):
            for k, v in value.items():
                self.delta(f"{ keys_with_dots }.{ k }", v)
        else:
            if self.get(keys_with_dots):
                self.append(keys_with_dots, value)
            else:
                self.set(keys_with_dots, value)

    def append(self, keys_with_dots: str, value: any):
        pointer, key = self.__locate_pointer(keys_with_dots)
        if isinstance(pointer[key], list):
            pointer[key].append(value)
        elif pointer[key] == {}:
            pointer[key] = []
            pointer[key].append(value)
        else:
            pointer[key] = [pointer[key]]
            pointer[key].append(value)
        return self

    def extend(self, keys_with_dots: str, value: any):
        if not isinstance(value, list):
            value = [value]
        pointer, key = self.__locate_pointer(keys_with_dots)
        if isinstance(pointer[key], list):
            pointer[key].extend(value)
        elif pointer[key] == {}:
            pointer[key] = []
            pointer[key].extend(value)
        else:
            pointer[key] = [pointer[key]]
            pointer[key].extend(value)
        return self 

    def __update_dict(self, pointer: dict, pointer_key: str, dict_item: any):
        if isinstance(dict_item, dict):
            for key in dict_item.keys():
                if key not in pointer[pointer_key]:
                    pointer[pointer_key][key] = {}
                if isinstance(dict_item[key], dict):
                    self.__update_dict(pointer[pointer_key], key, dict_item[key])
                else:
                    if dict_item[key] != None:
                        pointer[pointer_key][key] = dict_item[key]
        else:
            if dict_item != None:
                pointer[pointer_key] = dict_item

    def update(self, keys_with_dots: str, value: any):
        pointer, key = self.__locate_pointer(keys_with_dots)
        self.__update_dict(pointer, key, value)
        return self

    def update_by_dict(self, data_dict: dict):
        for key, value in data_dict.items():
            self.update(key, value)
        return self

    def _deep_get(self, data):
        result = None
        if isinstance(data, dict):
            result = {}
            for key, value in data.items():
                result.update({ key: self._deep_get(data[key]) })
            return result
        elif isinstance(data, list):
            result = []
            for item in data:
                result.append(self._deep_get(item))
            return result
        elif isinstance(data, set):
            result = set()
            for item in list(data):
                result.add(self._deep_get(item))
            return result
        elif isinstance(data, tuple):
            result = []
            for item in list(data):
                result.append(self._deep_get(item))
            return tuple(result)
        elif hasattr(data, '__class__') and hasattr(data, '__module__') and data.__module__ != 'builtins':
            return data
        else:
            return copy.deepcopy(data)

    def get(self, keys_with_dots: (str, None) = None, default: str=None, *, no_copy: bool = False):
        if self.target_data == None:
            self.target_data = self.get_target_data() or {}
        if keys_with_dots:
            keys = keys_with_dots.split('.')
            pointer = self.target_data
            for key in keys:
                if key not in pointer:
                    return default
                else:
                    pointer = pointer[key]
            if self.no_copy or no_copy:
                return pointer
            else:
                if hasattr(pointer, '__class__') and hasattr(pointer, '__module__') and pointer.__module__ != 'builtins':
                    return pointer
                else:
                    try:
                        return copy.deepcopy(pointer)
                    except TypeError:
                        return self._deep_get(pointer)
        else:
            if self.no_copy or no_copy:
                return self.target_data
            else:
                if hasattr(self.target_data, '__class__') and hasattr(self.target_data, '__module__') and self.target_data.__module__ != 'builtins':
                    return self.target_data
                else:
                    try:
                        return copy.deepcopy(self.target_data)
                    except TypeError:
                        return self.target_data

    def remove(self, keys_with_dots: str):
        pointer, key = self.__locate_pointer(keys_with_dots)
        del pointer[key]
        return self

    def empty(self):
        self.target_data = {}
        return self

    def save(self):
        final_data = self.get()
        return self.storage.set_all(self.table_name, final_data)

"""
class StorageEditor(DataOps):
    def __init__(self, target_data: (dict, None), *, storage: object, table_name: str):
        self.storage = storage
        self.table_name = table_name
        super().__init__(target_data = target_data)

    def save(self):
        final_data = self.get()
        return self.storage.set_all(self.table_name, final_data)
"""
        
class StorageDelegate(object):
    def __init__(self, *, db_name: str = "default", plugin_manager: object, settings: object):
        self.db_name = db_name
        self.plugin_manager = plugin_manager
        self.settings = settings
        return

    def set_storage_type(self, storage_type: str):
        self.settings.set("storage_type", storage_type)
        return self

    def __get_storage_plugin(self):
        storage_type = self.settings.get_trace_back("storage_type")
        return self.plugin_manager.get("storage", storage_type)(db_name = self.db_name, settings = self.settings)

    def set(self, table_name: str, key: str, value: any):
        self.__get_storage_plugin().set(table_name, key, value)
        return self

    def set_all(self, table_name: str, full_data: dict):
        self.__get_storage_plugin().set_all(table_name, full_data)
        return self

    def remove(self, table_name: str, key: str):
        self.__get_storage_plugin().remove(table_name, key)
        return self

    def update(self, table_name:str, update_data: dict):
        self.__get_storage_plugin().update(table_name, update_data)
        return self

    def get(self, table_name: str, key: str):
        return self.__get_storage_plugin().get(table_name, key)

    def get_all(self, table_name: str, keys: (list, None)=None):
        return self.__get_storage_plugin().get_all(table_name, keys)

    def table(self, table_name: str):
        get_table_data = lambda: self.get_all(table_name)
        return StorageEditor(get_target_data = get_table_data, storage = self, table_name = table_name)


from openai import AsyncOpenAI as OpenAIClient
from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, find_json, format_request_messages
from Agently.utils import RuntimeCtxNamespace
import httpx

class OAIClient(RequestABC):
    def __init__(self, request):
        self.request = request
        self.model_name = "OAIClient"
        self.model_settings = RuntimeCtxNamespace(f"model.{ self.model_name }", self.request.settings)
        self.request_type = self.request.request_runtime_ctx.get("request_type", "chat")
        self.default_options = {
            "stream": True
        }
        if not self.model_settings.get_trace_back("message_rules.no_multi_system_messages"):
            self.model_settings.set("message_rules.no_multi_system_messages", True)
        if not self.model_settings.get_trace_back("message_rules.strict_orders"):
            self.model_settings.set("message_rules.strict_orders", True)
        if not self.model_settings.get_trace_back("message_rules.no_multi_type_messages"):
            self.model_settings.set("message_rules.no_multi_type_messages", True)    

    def construct_request_messages(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[GENERAL INSTRUCTION]\n{ to_instruction(general_instruction_data) }" }] })
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[ROLE SETTINGS]\n{ to_instruction(role_data) }" }] })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            request_messages.append({ "role": "system", "content": [{"type": "text", "text": f"[ABOUT USER]\n{ to_instruction(user_info_data) }" }] })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "assistant", "content": [{"type": "text", "text": f"[ABSTRACT]\n{ to_instruction(headline_data) }" }] })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        prompt_text = ""
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            prompt_text = to_instruction(prompt_input_data)
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[HELPFUL INFORMATION]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[OUTPUT REQUIREMENT]"] = {
                        "TYPE": "JSON String can be parsed",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[OUTPUT REQUIERMENT]"] = str(prompt_output_data)
            prompt_text = to_prompt_structure(prompt_dict, end="[OUTPUT]:\n")
        request_messages.append({ "role": "user", "content": [{"type": "text", "text": prompt_text }] })
        return request_messages

    def construct_completion_prompt(self):
        # - init prompt
        completion_prompt = ""
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            completion_prompt += f"[GENERAL INSTRUCTION]\n{ to_instruction(general_instruction_data) }\n"
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            completion_prompt += f"[ROLE SETTINGS]\n{ to_instruction(role_data) }\n"
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            completion_prompt += f"[ABOUT USER]\n{ to_instruction(user_info_data) }\n"
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            completion_prompt += f"[ABSTRACT]\n{ to_instruction(headline_data) }\n"
        # - main prompt
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data and not chat_history_data:
            raise Exception("[Request] Missing 'prompt.chat_history', 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        # --- only input
        if prompt_input_data and not chat_history_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            if completion_prompt == "":
                completion_prompt += to_instruction(prompt_input_data)
            else:
                completion_prompt += f"[OUTPUT]\n{ to_instruction(prompt_input_data) }"
        # --- construct prompt
        else:
            # - with output format
            if isinstance(prompt_output_data, (dict, list, set)):
                prompt_dict = {}
                if chat_history_data:
                    prompt_dict["[HISTORY LOGS]"] = ""
                    for message in chat_history_data:
                        prompt_dict["[HISTORY LOGS]"] += f"{ message['role'] }: { message['content'] }\n"
                if prompt_input_data:
                    prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
                if prompt_info_data:
                    prompt_dict["[HELPFUL INFORMATION]"] = str(prompt_info_data)
                if prompt_instruct_data:
                    prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
                if prompt_output_data:
                    prompt_dict["[OUTPUT REQUIREMENT]"] = {
                        "TYPE": "JSON can be parsed in Python",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                completion_prompt += to_prompt_structure(
                    prompt_dict,
                    end="[OUTPUT]:\nassistant: "
                        if chat_history_data and len(chat_history_data) > 0
                        else "[OUTPUT]:\n"
                )
            # without output format
            else:
                if prompt_info_data:
                    completion_prompt += f"[HELPFUL INFORMATION]\n{ str(prompt_info_data) }\n"
                if prompt_instruct_data:
                    completion_prompt += f"[INSTRUCTION]\n{ to_instruction(prompt_instruct_data) }\n"
                if prompt_output_data:
                    completion_prompt += f"[OUTPUT REQUIREMENT]\n{ str(prompt_output_data) }\n"
                completion_prompt += "[OUTPUT]\n"
                # chat completion
                if chat_history_data:
                    for message in chat_history_data:
                        completion_prompt += f"{ message['role'] }: { message['content'] }\n"
                    if prompt_input_data:
                        completion_prompt += f"user: { to_instruction(prompt_input_data) }\n"
                    completion_prompt += "assistant: "
                # text completion
                else:
                    if prompt_input_data:
                        completion_prompt += prompt_input_data
        return completion_prompt

    def generate_request_data(self):
        # collect options
        options = self.model_settings.get_trace_back("options", {})
        for key, value in self.default_options.items():
            if key not in options:
                options.update({ key: value })
        if self.request_type == "chat":
            return {
                "messages": format_request_messages(self.construct_request_messages(), self.model_settings),
                **options,
            }
        elif self.request_type == "completions":
            return {
                "prompt": self.construct_completion_prompt(),
                **options,
            }
        else:
            raise Exception(f"Can not support request type: { self.request_type }.")

    async def request_model(self, request_data: dict):
        # create client
        client_params = {}
        base_url = self.model_settings.get_trace_back("url")
        if base_url:
            client_params.update({ "base_url": base_url })
        httpx_client = self.model_settings.get_trace_back("httpx_client")
        if httpx_client:
            httpx_client.headers.update({ "Connection": "close" })
            client_params.update({ "http_client": httpx_client })
        else:
            proxy = self.request.settings.get_trace_back("proxy")
            verify = self.model_settings.get_trace_back("verify")
            httpx_options = self.model_settings.get_trace_back("httpx.options", {})
            httpx_params = httpx_options
            httpx_params.update({
                "headers": [("Connection", "close")],
            })
            # verify
            if verify:
                httpx_params.update({ "verify": verify })
            # proxy
            if proxy:
                httpx_params.update({ "proxies": proxy })
            client_params.update({
                "http_client": httpx.AsyncClient(**httpx_params),
            })
        api_key = self.model_settings.get_trace_back("auth.api_key")
        if api_key:
            client_params.update({ "api_key": api_key })
        else:
            client_params.update({ "api_key": "None" })
        client = OpenAIClient(**client_params)
        # request
        if self.request_type == "chat":
            stream = await client.chat.completions.create(**request_data)
        elif self.request_type == "completions":
            stream = await client.completions.create(**request_data)
        return stream

    async def broadcast_response(self, response_generator):
        if self.request_type == "chat":
            response_message = {}
            async for part in response_generator:
                delta = dict(part.choices[0].delta)
                for key, value in delta.items():
                    if key not in response_message:
                        response_message[key] = value or ""
                    else:
                        response_message[key] += value or ""
                yield({ "event": "response:delta_origin", "data": part })
                yield({ "event": "response:delta", "data": part.choices[0].delta.content or "" })
            yield({ "event": "response:done_origin", "data": response_message })
            yield({ "event": "response:done", "data": response_message["content"] })
        elif self.request_type == "completions":
            response_message = {}
            async for part in response_generator:
                if "choices" in dir(part) and isinstance(part.choices, list) and len(part.choices) > 0:
                    delta = dict(part.choices[0])
                    for key, value in delta.items():
                        if key not in response_message and value is not None:
                            response_message[key] = value
                        elif key in response_message and value is not None:
                            response_message[key] += value
                    yield({ "event": "response:delta_origin", "data": delta })
                    yield({ "event": "response:delta", "data": delta["text"] or "" })
                else:
                    if self.request.settings.get_trace_back("is_debug"):
                        print(f"[Request] Server Response Message: { str(dict(part)) }")
                yield({ "event": "response:delta_origin", "data": part })
            yield({ "event": "response:done_origin", "data": response_message })
            yield({ "event": "response:done", "data": response_message["text"] })
                
    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }

def export():
    return("OAIClient", OAIClient)


from ..Schema import Schema
from ..lib.constants import EXECUTOR_TYPE_NORMAL, EXECUTOR_TYPE_START, DEFAULT_INPUT_HANDLE_VALUE
import uuid

def generate_executed_schema(schema: dict):
    """根据定义描述，生成执行描述"""
    edges_map_info = generate_edge_prepare_data(schema.get('edges') or [])
    all_chunks = schema.get('chunks') or []

    def create_exec_chunk(chunk):
        next_chunks = []
        next_chunk_map = {}
        for target in edges_map_info['edges_target_map'].get(chunk['id'], []):
            target_id = target['target']
            next_chunk = next_chunk_map.get(target_id)
            # 如之前没存入，则新建
            if not next_chunk:
                next_chunk = {
                    'id': target_id,
                    'handles': []
                }
                next_chunks.append(next_chunk)
                next_chunk_map[next_chunk['id']] = next_chunk
            # 往 chunk 中注入当前的 handle
            next_chunk['handles'].append({
                'handle': target['target_handle'],
                'source_handle': target['source_handle'],
                'condition': target.get('condition') or None
            })
                
        hanldes_desc = chunk.get('handles') or {}
        inputs_desc = hanldes_desc.get('inputs') or []
        if not len(inputs_desc):
            inputs_desc = [{'handle': DEFAULT_INPUT_HANDLE_VALUE}]
        return {
            'id': chunk['id'],
            'loop_entry': False, # 是否是循环的起点
            'next_chunks': next_chunks,
            'type': chunk.get('type') or EXECUTOR_TYPE_NORMAL,
            'executor': chunk.get('executor'),
            'deps': [
                {
                    'handle': input_desc.get('handle'),
                    # 运行时的依赖值，会随运行时实时更新，初始尝试从定义中取默认值
                    'data_slots': [{
                        'id': str(uuid.uuid4()), # 唯一 id，设置后就不变了，用于标识管理
                        'is_ready': True,
                        'execution_ticket': str(uuid.uuid4()), # 可执行票据，可派发
                        'updator': 'default', # 更新者
                        'value': input_desc.get('default')
                    }] if (input_desc.get('default') != None) or (chunk.get('type') == EXECUTOR_TYPE_START) else []
                }
                for input_desc in inputs_desc
            ],
            'data': chunk
        }
    
    runtime_chunks_map = {}
    for chunk in all_chunks:
        runtime_chunks_map[chunk['id']] = create_exec_chunk(chunk)
    
    # 标识循环起点
    def update_loop_walker(chunk, paths = []):
        for next_info in chunk.get('next_chunks'):
            next_chunk = runtime_chunks_map.get(next_info['id'])
            if next_chunk:
                if next_chunk['id'] in paths:
                    next_chunk['loop_entry'] = True
                    return
                update_loop_walker(next_chunk, paths + [next_chunk['id']])

    entries = []
    for chunk in all_chunks:
        if not edges_map_info['edges_source_map'].get(chunk['id'], []):
            entry_chunk = runtime_chunks_map.get(chunk['id'])
            if entry_chunk:
                entries.append(entry_chunk)
                update_loop_walker(entry_chunk, [entry_chunk['id']])

    return {
        'entries': entries,
        'chunk_map': runtime_chunks_map
    }

def create_empty_data_slot(chunk):
    return {
        'is_ready': chunk.get('type') == EXECUTOR_TYPE_START,
        'updator': '',
        'value': None
    }

def create_empty_data_slots(chunk):
    return [{
        'is_ready': True,
        'updator': '',
        'value': None
    }] if chunk.get('type') == EXECUTOR_TYPE_START else []

def reset_chunk_slot_val(slot):
    """清空掉某个dep slot 的值"""
    slot['value'] = None
    slot['is_ready'] = False
    slot['execution_ticket'] = ''
    slot['updator'] = ''

def disable_chunk_dep_ticket(slot):
    """回收执行票据"""
    slot['execution_ticket'] = ''

def create_new_chunk_slot_with_val(ref_chunk_id, new_val):
    """创建一个新的带有值的slot"""
    return {
        'id': str(uuid.uuid4()), # 唯一 id，设置后就不变了，用于标识管理
        'is_ready': True,
        'execution_ticket': str(uuid.uuid4()), # 可执行票据，可派发
        'updator': ref_chunk_id, # 更新者
        'value': new_val
    }

def generate_edge_prepare_data(edges: list):
    edges_target_map = {}
    edges_source_map = {}

    for edge in edges:
        edges_target_map.setdefault(edge['source'], []).append(edge)
        edges_source_map.setdefault(edge['target'], []).append(edge)

    return {
        'edges_source_map': edges_source_map,
        'edges_target_map': edges_target_map
    }



from .utils import RequestABC, to_instruction, to_json_desc, to_prompt_structure
from Agently.utils import RuntimeCtxNamespace
import zhipuai

class ZhipuAI(RequestABC):
    def __init__(self, request):
        self.request = request
        self.model_name = "ZhipuAI"
        self.model_settings = RuntimeCtxNamespace(f"model.{ self.model_name }", self.request.settings)
        self.request_type = self.request.request_runtime_ctx.get("request_type", "chat")
        if self.request_type == None:
            self.request_type = "chat"

    def construct_request_messages(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "user", "content": f"[重要指导说明]\n{ to_instruction(general_instruction_data) }" })
            request_messages.append({ "role": "assistant", "content": "OK" })
        # - role
        role_data = self.request.request_runtime_ctx.get("prompt.role")
        if role_data and self.request_type == "chat":
            request_messages.append({ "role": "user", "content": f"[角色及行为设定]\n{ to_instruction(role_data) }" })
            request_messages.append({ "role": "assistant", "content": "OK" })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get("prompt.user_info")
        if user_info_data and self.request_type == "chat":
            request_messages.append({ "role": "user", "content": f"[用户信息]\n{ to_instruction(user_info_data) }" })
            request_messages.append({ "role": "assistant", "content": "OK" })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "user", "content": f"[主题及摘要]{to_instruction(headline_data)}" })
            request_messages.append({ "role": "assistant", "content": "OK" })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get("prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            request_messages.append({ "role": "user", "content": to_instruction(prompt_input_data) })
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[输入]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[补充信息]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[处理规则]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[输出要求]"] = {
                        "TYPE": "JSON can be parsed in runtime",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[输出要求]"] = str(prompt_output_data)
            request_messages.append({ "role": "user", "content": to_prompt_structure(prompt_dict, end="[输出]:\n") })
        return request_messages

    def construct_request_messages_for_gml4(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "system", "content": f"[重要指导说明]\n{ to_instruction(general_instruction_data) }" })
        # - role
        role_data = self.request.request_runtime_ctx.get("prompt.role")
        if role_data:
            request_messages.append({ "role": "system", "content": f"[角色及行为设定]\n{ to_instruction(role_data) }" })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get("prompt.user_info")
        if user_info_data:
            request_messages.append({ "role": "system", "content": f"[用户信息]\n{ to_instruction(user_info_data) }" })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "system", "content": f"[主题及摘要]{to_instruction(headline_data)}" })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get("prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            request_messages.append({ "role": "user", "content": to_instruction(prompt_input_data) })
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[输入]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[补充信息]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[处理规则]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[输出要求]"] = {
                        "TYPE": "JSON can be parsed in runtime",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[输出要求]"] = str(prompt_output_data)
            request_messages.append({ "role": "user", "content": to_prompt_structure(prompt_dict, end="[输出]:\n") })
        return request_messages

    def generate_request_data(self):
        prompt = None
        messages = None
        options = self.model_settings.get_trace_back("options", {})
        if self.request_type == "chat" and "model" not in options:
            options["model"] = "glm-4"
        if self.request_type == "chat" and options["model"] == "glm-3-turbo":
            messages = self.construct_request_messages()
            options["stream"] = True
            return {
                "messages": messages,
                **options
            } 
        elif self.request_type == "chat" and options["model"] == "glm-4":
            messages = self.construct_request_messages_for_gml4()
            options["stream"] = True
            return {
                "messages": messages,
                **options
            } 
        elif self.request_type == "character":
            options["model"] = "characterglm"
            prompt = self.construct_request_messages()
            options["incremental"] = True
            role_data = self.request.request_runtime_ctx.get("prompt.role")
            options["meta"] = {}
            if role_data:
                if "NAME" in role_data:
                    options["meta"]["bot_name"] = role_data["NAME"][0]
                    del role_data["NAME"]
                else:
                    options["meta"]["bot_name"] = "智能助理"
                options["meta"]["bot_info"] = to_instruction(role_data)
            else:
                options["meta"]["bot_name"] = "智能助理"
                options["meta"]["bot_info"] = "由[Agently智能体开发框架](Agently.cn)创建的智能助理"
            user_info_data = self.request.request_runtime_ctx.get("prompt.user_info")
            if user_info_data:
                if "NAME" in user_info_data:
                    options["meta"]["user_name"] = user_info_data["NAME"][0]
                    del user_info_data["NAME"]
                else:
                    options["meta"]["user_name"] = "用户"
                options["meta"]["user_info"] = to_instruction(user_info_data)
            else:
                options["meta"]["user_name"] = "用户"
                options["meta"]["user_info"] = "用户"
            if len(options["meta"].keys()) == 0:
                del options["meta"]
            return {
                "prompt": prompt,
                **options
            } 
        elif self.request_type == "embedding":
            options["model"] = "text_embedding"
            prompt = self.request.request_runtime_ctx.get("prompt.input")
            return {
                "prompt": prompt,
                **options
            } 
        else:
            raise Exception(f"[Request] ZhipuAI: can not request model with this options, please check:\n{ str(options) }")

    async def request_model(self, request_data: dict):
        zhipuai.api_key = self.model_settings.get_trace_back("auth.api_key")
        if self.request_type == "chat":
            client = zhipuai.ZhipuAI(api_key = zhipuai.api_key)
            return client.chat.completions.create(**request_data)
        if self.request_type == "character":
            return zhipuai.model_api.sse_invoke(**request_data)
        else:
            return zhipuai.model_api.invoke(**request_data)

    async def broadcast_response(self, response_generator):
        if self.request_type == "chat":
            buffer = ""
            for chunk in response_generator:
                if chunk.choices[0].delta:
                    if not chunk.choices[0].finish_reason:
                        delta = chunk.choices[0].delta.content
                        yield({ "event": "response:delta", "data": delta })
                        yield({ "event": "response:delta_origin", "data": dict(chunk) })
                        buffer += delta
                    else:
                        yield({ "event": "response:done", "data": buffer })
                        yield({ "event": "response:done_origin", "data": { "content": buffer, "meta": dict(chunk) } })
        elif self.request_type == "character":
            buffer = ""
            for part in response_generator.events():
                if part.event == "add":
                    yield({ "event": "response:delta", "data": part.data })
                    buffer += part.data
                elif part.event == "error" or part.event == "interrupted":
                    raise Exception(f"[Request] ZhipuAI Error: { part.data }")
                elif part.event == "finish":
                    yield({ "event": "response:done", "data": buffer })
                    yield({ "event": "response:done_origin", "data": { "content": buffer, "meta": part.meta } })
        elif self.request_type == "embedding":
            yield({ "event": "response:done", "data": response_generator["data"]["embedding"] })

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }

def export():
    return ("ZhipuAI", ZhipuAI)


def format_request_messages(request_messages, model_settings):
    system_prompt = ""
    system_messages = []
    chat_messages = []
    role_list = ["user", "assistant"]
    current_role = 0
    for message in request_messages:
        if "role" not in message or "content" not in message:
            raise Exception(f"[Agently Request]: key 'role' and key 'content' must be contained in `request_messages`:\n{ str(request_messages) }")
        if message["role"] == "system":
            # no_multi_system_messages=True
            if model_settings.get_trace_back("message_rules.no_multi_system_messages"):
                if isinstance(message["content"], str):
                    system_prompt += f"{ message['content'] }\n"
                elif isinstance(message["content"], list):
                    for content in message["content"]:
                        if "type" in content and content["type"] == "text":
                            system_prompt += f"{ content['text'] }\n"
                        else:
                            system_prompt += f"{ str(content) }\n"
                else:
                    system_prompt += f"{ str(message['content']) }\n"
            # no_multi_system_messages=False
            else:
                system_messages.append(message)
        else:
            # strict_orders=True
            if model_settings.get_trace_back("message_rules.strict_orders"):
                if len(chat_messages) == 0 and message["role"] != "user":
                    chat_messages.append({ "role": "user", "content": [{"type": "text", "text": "What did we talked about?" }] })
                    current_role = not current_role
                if message["role"] == role_list[current_role]:
                    chat_messages.append(message)
                    current_role = not current_role
                else:
                    content = f"{ chat_messages[-1]['content'] }\n{ message['content'] }"
                    chat_messages[-1]['content'] = content
            # strict_orders=False
            else:
                chat_messages.append(message)
    # no_multi_system_messages=True
    if model_settings.get_trace_back("message_rules.no_multi_system_messages") and system_prompt != "":
        system_messages.append({
            "role": "system",
            "content": [{
                "type": "text",
                "text": system_prompt
            }]
        })
    formatted_messages = system_messages.copy()
    formatted_messages.extend(chat_messages)
    # no_multi_type_messages=True
    if model_settings.get_trace_back("message_rules.no_multi_type_messages"):
        current_messages = formatted_messages.copy()
        formatted_messages = []
        for message in current_messages:
            if isinstance(message["content"], str):
                formatted_messages.append(message)
            elif isinstance(message["content"], list):
                for content in message["content"]:
                    if "type" in content and content["type"] == "text":
                        formatted_messages.append({ "role": message["role"], "content": content["text"] })
                    else:
                        formatted_messages.append({ "role": message["role"], "content": str(content) })
            else:
                formatted_messages.append({ "role": message["role"], "content": str(content) })
    return formatted_messages


import os
import time
import subprocess

def streamlit_app(app_connector, **kwargs):
    is_multi_round = kwargs.get("multi_round", True)
    is_fluency = kwargs.get("fluency", False)
    st = app_connector.runtime.get("package.streamlit")
    if not st:
        try:
            import streamlit as st
            app_connector.runtime.set("package.streamlit", st)
        except:
            raise Exception("[Application Connector] Package 'streamlit' require to be installed. Use 'pip install streamlit' to install.")
    def get_chat_response():
        for item in app_connector.data_generator.start():
            if is_fluency:
                for char in list(item):
                    time.sleep(0.01)
                    yield char
            else:
                yield item
    st.title("A Chatbot powered by Agently & Streamlit")
    if "chat_history" not in st.session_state:
        st.session_state.chat_history = []
    for message in st.session_state.chat_history:
        with st.chat_message(message["role"]):
            st.markdown(message["content"])
    message = st.chat_input("Now let's talk...")
    if message:
        if not app_connector.message_handler:
            st.warning("No Message Handler, use `.set_message_handler()` to set one.", icon="⚠️")
        
        # Refresh binded agent event
        app_connector.refresh_binded_agent()
        # Define generator event handlers
        app_connector.on("delta", lambda data: { "yield": data })
        app_connector.on("done", lambda data: { "end": True })
        with st.chat_message("user"):
            st.markdown(message)
        st.session_state.chat_history.append({"role": "user", "content": message})
        with st.spinner("Processing..."):
            with st.chat_message("assisitant"):
                if is_multi_round:
                    app_connector.run_message_handler(
                        message,
                        st.session_state.chat_history[:-1]
                        if len(st.session_state.chat_history) > 0
                        else []
                    )
                else:
                    app_connector.run_message_handler(message, [])
                response_generator = get_chat_response()
                response = st.write_stream(response_generator)
                st.session_state.chat_history.append({ "role": "assistant", "content": response })


"""
import copy

class Store:
  def __init__(self, init_schema = {}) -> None:
    self.store = init_schema.copy()

  def set(self, name: str, data: any):
    self.store[name] = data
  
  def set_with_dict(self, data: dict):
    self.store.update(data)
  
  def remove(self, name: str):
    if name in self.store:
      del self.store[name]
    return self
  
  def remove_all(self):
    self.store = {}
    return self

  def get(self, name: str, default = None):
    return self.store.get(name) or default
  
  def get_all(self):
    return copy.deepcopy(self.store)
"""

import Agently.utils.DataOps as DataOps

class Store(DataOps):
  def __init__(self, init_schema = None):
    super().__init__(target_data = init_schema)
    self.store = self.target_data
    self.remove_all = self.empty
    self.set_with_dict = self.update_by_dict
    self.get_all = lambda: self.get()


from .RuntimeCtx import RuntimeCtx, RuntimeCtxNamespace
from .StorageDelegate import StorageDelegate
from .PluginManager import PluginManager
from .AliasManager import AliasManager
from .ToolManager import ToolManager
from .IdGenerator import IdGenerator
from .DataOps import DataOps, NamespaceOps
from .transform import to_prompt_structure, to_json_desc, to_instruction, find_all_jsons, find_json
from .check_version import check_version
from .load_json import load_json, find_and_load_json
from .DataGenerator import DataGenerator



import inspect
from ...lib.Store import Store
from ...lib.constants import DEFAULT_INPUT_HANDLE_VALUE

def use_loop_executor(sub_workflow):
    async def loop_executor(inputs, store: Store, **sys_info):
        # Run Loop
        input_val = inputs.get(DEFAULT_INPUT_HANDLE_VALUE)
        all_result = []
        if isinstance(input_val, list):
            for val in input_val:
                all_result.append(await loop_unit_core(unit_val=val, store=store))
        elif isinstance(input_val, dict):
            for key, value in input_val.items():
                all_result.append(await loop_unit_core(unit_val={
                    "key": key,
                    "value": value
                }, store=store))
        elif isinstance(input_val, int):
            for i in range(input_val):
                all_result.append(await loop_unit_core(unit_val=i, store=store))
        # Old Version Compatible
        if inspect.isfunction(sub_workflow) or inspect.iscoroutinefunction(sub_workflow):
            return all_result
        elif hasattr(sub_workflow, "settings") and sub_workflow.settings.get("compatible_version") and sub_workflow.settings.get("compatible_version") <= 3316:
            return all_result
        else:
            # Regroup
            final_result = {}
            i = 0
            for one_result in all_result:
                if one_result:
                    used_handle_pool = []
                    for handle, result in one_result.items():
                        if handle not in final_result:
                            final_result.update({ handle: [None] * i })
                        final_result[handle].append(result)
                        used_handle_pool.append(handle)
                    for final_result_handle in final_result.keys():
                        if final_result_handle not in used_handle_pool:
                            final_result[final_result_handle].append(None)
                else:
                    if "default" not in final_result:
                        final_result.update({ "default": [None] * i })
                    final_result["default"].append(None)
                    for final_result_handle in final_result.keys():
                        if final_result_handle != "default":
                            final_result[final_result_handle].append(None)
                i += 1
            return final_result

    async def loop_unit_core(unit_val, store):
        if inspect.iscoroutinefunction(sub_workflow):
            return await sub_workflow(unit_val, store)
        elif inspect.isfunction(sub_workflow):
            return sub_workflow(unit_val, store)
        else:
            return await sub_workflow.start_async(unit_val)
    
    return loop_executor



#ABC = Abstract Base Class
from Agently.utils import to_instruction, to_json_desc
from .transform import to_prompt_structure
from abc import ABC, abstractmethod

class RequestABC(ABC):
    def __init__(self, request):
        self.request = request
        self.model_name = "<Model-Name>"
        # if you want to use RuntimeCtxNamespace to help you manage your settings
        # from Agently.utils import RuntimeCtxNamespace
        # self.model_settings = RuntimeCtxNamespace(f"model.{ self.model_name }", self.request.settings)

    def construct_request_messages(self):
        '''
        This is an example to help you understand how to construct your own request message or prompt
        Here're Standard LLM Request Inputs from Agently that you can use :
        self.request.request_runtime_ctx.get("prompt.system")
        self.request.request_runtime_ctx.get("prompt.abstract")
        self.request.request_runtime_ctx.get("prompt.chat_history")
        self.request.request_runtime_ctx.get("prompt.input")
        self.request.request_runtime_ctx.get("prompt.info")
        self.request.request_runtime_ctx.get("prompt.instruct")
        self.request.request_runtime_ctx.get("prompt.output")
        You can use them to construct request or prompt following the model rules.
        And down below is just one example that usually fits OpenAI messages rules.
        '''
        #init request messages
        request_messages = []
        # - system message
        system_data = self.request.request_runtime_ctx.get("prompt.system")
        if system_data:
            request_messages.append({ "role": "system", "content": to_instruction(system_data) })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "assistant", "content": to_instruction(headline_data) })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get("prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            request_messages.append({ "role": "user", "content": to_instruction(prompt_input_data) })
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[HELPFUL INFORMATION]"] = to_instruction(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[OUTPUT REQUIREMENT]"] = {
                        "TYPE": "JSON can be parsed in Python",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[OUTPUT REQUIERMENT]"] = str(prompt_output_data)
            request_messages.append({ "role": "user", "content": to_prompt_structure(prompt_dict, end="[OUTPUT]:\n") })
        return request_messages

    def generate_request_data(self):
        # ...
        return {
            #Your request data dict
            #"stream": True,
            #"messages": [...],
        }

    def request_model(self, request_data: dict):
        return
            #response generator / response result

    def broadcast_response(self, response_generator):
        '''
        Use response_generator or response result to broadcast event data using yield
        Standard Event Data Format:
        * { "event": "response:delta", "data": any } (optional, important for streaming output)
        * { "event": "response:done", "data": any } (required)
        - { "event": "response:delta_origin", "data": any } (optional)
        - { "event": "response:done_origin", "data": any } (optional)
        If your response is not streaming, you can only yield response:done.
        '''
        return

    @abstractmethod
    def export(self):
        return {
            "generate_request_data": callable,# (get_settings, request_runtime_ctx) -> request_data: dict
            "request_model": callable,# (request_data) -> response_generator
            "broadcast_response": callable,# (response_generator) -> broadcast_event_generator
        }


from .utils import ToolABC

class Code(ToolABC):
    def __init__(self, tool_manager: object):
        self.tool_manager = tool_manager

    def calculate(self, code: str, result_vars: list=None):
        results = {
            "origin": None,
            "for_agent": None,
        }
        global_vars = self.tool_manager.runtime_cache.get("global_vars", {})
        local_vars = self.tool_manager.runtime_cache.get("local_vars", {})
        exec(code, global_vars, local_vars)
        self.tool_manager.runtime_cache.update({ "global_vars": global_vars })
        self.tool_manager.runtime_cache.update({ "local_vars": local_vars })
        result = {}
        if result_vars:
            for result_var in result_vars:
                result.update({ result_var["why"]: local_vars.get(result_var["var_name"]) })
            results = {
                "origin": result,
                "for_agent": result,
            }
        else:
            results = {
                "origin": local_vars,
                "for_agent": local_vars,
            }
        return results

    def export(self):
        return {
            "calculate": {
                "desc": "execute Python code to storage result to variables",
                "args": {
                    "code": ("Python Code assigning values to variables", "[*Required]"),
                    "result_vars": [{
                        "var_name": ("String", "[*Required]var name in {code}"),
                        "why": ("String", "[*Required]brief purpose to get result of {var_name}")
                    }],
                },
                "func": self.calculate,
            },
        }

def export():
    return ("Code", Code)


import functools


def exposed_interface(type='command'):
    """修饰对外提供的接口调用，支持传参"""
    def decorator(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            # 调用所有前置拦截方法
            for interceptor in self._pre_command_interceptors:
                interceptor(self, type, method.__name__, *args, **kwargs)

            # 调用实际的方法
            result = method(self, *args, **kwargs)

            # 调用所有后置拦截方法
            for interceptor in self._post_command_interceptors:
                interceptor(self, type, method.__name__, result)

            return result
        return wrapper
    return decorator


from abc import ABC, abstractmethod
from typing import Dict, Any

class ChunkExecutorABC(ABC):
    def __init__(self, chunk_desc: dict, main_executor):
        self.chunk = chunk_desc
        self.main_executor = main_executor
        self.type = 'unknow'

    @abstractmethod
    def exec(self, inputs_with_handle_name: dict) -> Dict[str, Any]:
        """
        执行函数，要求返回 dict 结果，字段要求如下：
        1. status: 必须，表示执行状态，取值为 'success' 或 'error'
        2. dataset: 必须，执行结果，为 any 类型
        3. error_msg: 可选，错误信息，为 str 类型
        """
        pass



import time

def gradio_app(app_connector, **kwargs):
    """
    Support Kwargs:
        multi_round (bool=True): Turn on or turn off multi round chat.
        interface (dict): Additional interface options dict.
        launch (dict): Additional launch options dict.
    """
    is_multi_round = kwargs.get("multi_round", True)
    is_fluency = kwargs.get("fluency", False)
    try:
        import gradio as gr
    except:
        raise Exception("[Application Connector] Package 'gradio' require to be installed. Use 'pip install gradio' to install.")
    def app_executor(message, history):
        if app_connector.message_handler:
            # Refresh binded agent event
            app_connector.refresh_binded_agent()
            # Format chat history
            chat_history = []
            if is_multi_round:
                for round in history:
                    if round[0]:
                        chat_history.append({ "role": "user", "content": round[0] })
                    if round[1]:
                        chat_history.append({ "role": "assistant", "content": round[1] })
            # Define generator event handlers
            app_connector.on("buffer", lambda data: { "yield": data })
            app_connector.on("done", lambda data: { "yield": data, "end": True })
            # Run message handler
            app_connector.run_message_handler(message, chat_history)
            # Get response from data generator
            try:
                for item in app_connector.data_generator.start():
                    if is_fluency:
                        time.sleep(0.01)
                    if item:
                        yield item
                return
            except StopIteration:
                return
        else:
            return "No Message Handler, use `.set_message_handler()` to set one."
    gr.ChatInterface(
        app_executor,
        title="Agently Gradio Chatbot",
        description="A Chatbot powered by Agently & Gradio",
        textbox=gr.Textbox(placeholder="Now let's talk..."),
        **(kwargs.get("interface", {}))
    ).launch(**(kwargs.get("launch", {})))